/********************************************************************/
/*                                                                  */
/*  cmd_unx.c     Command functions which call the Unix API.        */
/*  Copyright (C) 1989 - 2016, 2018, 2019  Thomas Mertes            */
/*                                                                  */
/*  This file is part of the Seed7 Runtime Library.                 */
/*                                                                  */
/*  The Seed7 Runtime Library is free software; you can             */
/*  redistribute it and/or modify it under the terms of the GNU     */
/*  Lesser General Public License as published by the Free Software */
/*  Foundation; either version 2.1 of the License, or (at your      */
/*  option) any later version.                                      */
/*                                                                  */
/*  The Seed7 Runtime Library is distributed in the hope that it    */
/*  will be useful, but WITHOUT ANY WARRANTY; without even the      */
/*  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR */
/*  PURPOSE.  See the GNU Lesser General Public License for more    */
/*  details.                                                        */
/*                                                                  */
/*  You should have received a copy of the GNU Lesser General       */
/*  Public License along with this program; if not, write to the    */
/*  Free Software Foundation, Inc., 51 Franklin Street,             */
/*  Fifth Floor, Boston, MA  02110-1301, USA.                       */
/*                                                                  */
/*  Module: Seed7 Runtime Library                                   */
/*  File: seed7/src/cmd_unx.c                                       */
/*  Changes: 2010, 2012  Thomas Mertes                              */
/*  Content: Command functions which call the Unix API.             */
/*                                                                  */
/********************************************************************/

#define LOG_FUNCTIONS 0
#define VERBOSE_EXCEPTIONS 0

#include "version.h"

#define _XOPEN_SOURCE 500
#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "sys/types.h"
#include "sys/stat.h"
#include "fcntl.h"
#include "signal.h"
#include "errno.h"

#if UNISTD_H_PRESENT
#include "unistd.h"
#endif

#if HAS_GETGRGID_R || HAS_GETGRGID
#include "grp.h"
#endif

#if HAS_GETPWUID_R || HAS_GETPWUID
#include "pwd.h"
#endif

#include "common.h"
#include "data_rtl.h"
#include "os_decls.h"
#include "heaputl.h"
#include "striutl.h"
#include "arr_rtl.h"
#include "cmd_rtl.h"
#include "str_rtl.h"
#include "int_rtl.h"
#include "rtl_err.h"

#undef EXTERN
#define EXTERN
#include "cmd_drv.h"


#ifndef PATH_MAX
#define PATH_MAX 2048
#endif

#ifdef EMULATE_ENVIRONMENT
char **environ7 = NULL;
#endif



/**
 *  Get the absolute path of the executable of the current process.
 *  @param arg_0 Parameter argv[0] from the function main() as string.
 *  @return the absolute path of the current process.
 */
striType getExecutablePath (const const_striType arg_0)

  {
#if HAS_SYMBOLIC_LINKS
    os_charType buffer[PATH_MAX];
    ssize_t readlink_result;
    errInfoType err_info = OKAY_NO_ERROR;
#ifdef APPEND_EXTENSION_TO_EXECUTABLE_PATH
    striType exeExtension;
#endif
#endif
    striType cwd;
    striType executablePath;

  /* getExecutablePath */
    logFunction(printf("getExecutablePath\n"););
#if HAS_SYMBOLIC_LINKS
    readlink_result = readlink("/proc/self/exe", buffer, sizeof(buffer) - 1);
    if (readlink_result != -1) {
      buffer[readlink_result] = '\0';
      executablePath = cp_from_os_path(buffer, &err_info);
      if (unlikely(executablePath == NULL)) {
        logError(printf("getExecutablePath: cp_from_os_path failed.\n"););
        raise_error(err_info);
#ifdef APPEND_EXTENSION_TO_EXECUTABLE_PATH
      } else {
        exeExtension = CSTRI_LITERAL_TO_STRI(EXECUTABLE_FILE_EXTENSION);
        strAppendTemp(&executablePath, exeExtension);
#endif
      } /* if */
    } else {
#endif
      if (strChPos(arg_0, (charType) '/') == 0) {
        executablePath = examineSearchPath(arg_0);
      } else if (arg_0->size >= 1 && arg_0->mem[0] == (charType) '/') {
        executablePath = strCreate(arg_0);
      } else {
        cwd = cmdGetcwd();
        executablePath = concatPath(cwd, arg_0);
        FREE_STRI(cwd, cwd->size);
      } /* if */
      if (unlikely(executablePath == NULL)) {
        raise_error(MEMORY_ERROR);
#if HAS_SYMBOLIC_LINKS
      } else {
        executablePath = followLink(executablePath);
#endif
      } /* if */
#if HAS_SYMBOLIC_LINKS
    } /* if */
#endif
    logFunction(printf("getExecutablePath --> \"%s\"\n",
                       striAsUnquotedCStri(executablePath)););
    return executablePath;
  } /* getExecutablePath */



#ifdef EMULATE_ENVIRONMENT
char *getenv7 (const char *name)

  {
    size_t len;
    char **p, *c;

  /* getenv7 */
    if (name == NULL || environ7 == NULL || strchr(name, '=') != NULL) {
      return NULL;
    } else {
      len = strlen(name);
      for (p = environ7; (c = *p) != NULL; ++p) {
        if (strncmp(c, name, len) == 0 && c[len] == '=') {
          return &c[len + 1];
        } /* if */
      } /* for */
      return NULL;
    } /* if */
  } /* getenv7 */



int setenv7 (const char *name, const char *value, int overwrite)

  {
    size_t len;
    char **p, *c;
    size_t nameCount = 0;

  /* setenv7 */
    if (name == NULL || name[0] == '\0' || strchr(name, '=') != NULL) {
      errno = EINVAL;
      return -1;
    } else {
      len = strlen(name);
      if (environ7 != NULL) {
        for (p = environ7; (c = *p) != NULL; ++p) {
          nameCount++;
          if (strncmp(c, name, len) == 0 && c[len] == '=') {
            if (overwrite) {
              if ((*p = realloc(*p, len + strlen(value) + 2)) == NULL) {
                errno = ENOMEM;
                return -1;
              } else {
                strcpy(&c[len + 1], value);
              } /* if */
            } /* if **/
            return 0;
          } /* if */
        } /* for */
      } /* if */
      if ((environ7 = realloc(environ7, sizeof(char *) * (nameCount + 2))) == NULL ||
          (c = malloc(len + strlen(value) + 2)) == NULL) {
        errno = ENOMEM;
        return -1;
      } else {
        memcpy(c, name, len);
        c[len] = '=';
        strcpy(&c[len + 1], value);
        environ7[nameCount] = c;
        environ7[nameCount + 1] = NULL;
        return 0;
      } /* if */
    } /* if */
  } /* setenv7 */



int unsetenv7 (const char *name)

  {
    size_t len;
    char **p, *c;
    char **found = NULL;
    size_t nameCount = 0;

  /* unsetenv7 */
    if (name == NULL || name[0] == '\0' || strchr(name, '=') != NULL) {
      errno = EINVAL;
      return -1;
    } else {
      len = strlen(name);
      if (environ7 != NULL) {
        for (p = environ7; (c = *p) != NULL; ++p) {
          nameCount++;
          if (found == NULL && strncmp(c, name, len) == 0 && c[len] == '=') {
            found = p;
          } /* if */
        } /* for */
        if (found != NULL) {
          if ((environ7 = realloc(environ7, sizeof(char *) * (nameCount))) == NULL) {
            errno = ENOMEM;
            return -1;
          } else {
            free(*found);
            *found = environ7[nameCount - 1];
            environ7[nameCount - 1] = NULL;
          } /* if */
        } /* if */
      } /* if */
      return 0;
    } /* if */
  } /* unsetenv7 */
#endif



static striType getGroupFromGid (gid_t gid)

  {
#if HAS_GETGRGID_R || HAS_GETGRGID
#if HAS_GETGRGID_R
    int getgrgid_result;
    struct group grp;
    char buffer[2048];
#endif
    struct group *grpResult;
#endif
    striType group;

  /* getGroupFromGid */
#if HAS_GETGRGID_R || HAS_GETGRGID
#if HAS_GETGRGID_R
    getgrgid_result = getgrgid_r(gid, &grp, buffer,
                                 sizeof(buffer), &grpResult);
    if (unlikely(getgrgid_result != 0)) {
      logError(printf("getGroupFromGid: getgrgid_r(" FMT_U32 ", ...) failed:\n"
                      "errno=%d\nerror: %s\n",
                      gid, errno, strerror(errno)););
      raise_error(FILE_ERROR);
      group = NULL;
    }
#else
    grpResult = getgrgid(gid);
    if (unlikely(grpResult == NULL)) {
      logError(printf("getGroupFromGid: getgrgid(" FMT_U32 ") failed:\n"
                      "errno=%d\nerror: %s\n",
                      gid, errno, strerror(errno)););
      raise_error(FILE_ERROR);
      group = NULL;
    }
#endif
    else if (grpResult != NULL) {
      group = cstri8_or_cstri_to_stri(grpResult->gr_name);
      if (unlikely(group == NULL)) {
        raise_error(MEMORY_ERROR);
      } /* if */
    } else {
      group = intStr((intType) gid);
    } /* if */
#else
    group = intStr((intType) gid);
#endif
    return group;
  } /* getGroupFromGid */



static striType getUserFromUid (uid_t uid)

  {
#if HAS_GETPWUID_R || HAS_GETPWUID
#if HAS_GETPWUID_R
    int getpwuid_result;
    struct passwd pwd;
    char buffer[2048];
#endif
    struct passwd *pwdResult;
#endif
    striType user;

  /* getUserFromUid */
#if HAS_GETPWUID_R || HAS_GETPWUID
#if HAS_GETPWUID_R
    getpwuid_result = getpwuid_r(uid, &pwd, buffer,
                                 sizeof(buffer), &pwdResult);
    if (unlikely(getpwuid_result != 0)) {
      logError(printf("getUserFromUid: getpwuid_r(" FMT_U32 ", ...) failed:\n"
                      "errno=%d\nerror: %s\n",
                      uid, errno, strerror(errno)););
      raise_error(FILE_ERROR);
      user = NULL;
    }
#else
    pwdResult = getpwuid(uid);
    if (unlikely(pwdResult == NULL)) {
      logError(printf("getUserFromUid: getpwuid(" FMT_U32 ") failed:\n"
                      "errno=%d\nerror: %s\n",
                      uid, errno, strerror(errno)););
      raise_error(FILE_ERROR);
      user = NULL;
    }
#endif
    else if (pwdResult != NULL) {
      user = cstri8_or_cstri_to_stri(pwdResult->pw_name);
      if (unlikely(user == NULL)) {
        raise_error(MEMORY_ERROR);
      } /* if */
    } else {
      user = intStr((intType) uid);
    } /* if */
#else
    user = intStr((intType) uid);
#endif
    return user;
  } /* getUserFromUid */



striType cmdGetGroup (const const_striType filePath)

  {
    os_striType os_path;
    int path_info = PATH_IS_NORMAL;
    errInfoType err_info = OKAY_NO_ERROR;
    os_stat_struct stat_buf;
    int stat_result;
    striType group;

  /* cmdGetGroup */
    logFunction(printf("cmdGetGroup(\"%s\")", striAsUnquotedCStri(filePath));
                fflush(stdout););
    os_path = cp_to_os_path(filePath, &path_info, &err_info);
    if (unlikely(os_path == NULL)) {
      logError(printf("cmdGetGroup: cp_to_os_path(\"%s\", *, *) failed:\n"
                      "path_info=%d, err_info=%d\n",
                      striAsUnquotedCStri(filePath), path_info, err_info););
      raise_error(err_info);
      group = NULL;
    } else {
      stat_result = os_stat(os_path, &stat_buf);
      if (unlikely(stat_result != 0)) {
        logError(printf("cmdGetGroup: os_stat(\"" FMT_S_OS "\") failed:\n"
                        "errno=%d\nerror: %s\n",
                        os_path, errno, strerror(errno)););
        os_stri_free(os_path);
        raise_error(FILE_ERROR);
        group = NULL;
      } else {
        os_stri_free(os_path);
        /* printf("cmdGetGroup: st_gid=" FMT_U32 "\n", stat_buf.st_gid); */
        group = getGroupFromGid(stat_buf.st_gid);
      } /* if */
    } /* if */
    logFunctionResult(printf("\"%s\"\n", striAsUnquotedCStri(group)););
    return group;
  } /* cmdGetGroup */



striType cmdGetOwner (const const_striType filePath)

  {
    os_striType os_path;
    int path_info = PATH_IS_NORMAL;
    errInfoType err_info = OKAY_NO_ERROR;
    os_stat_struct stat_buf;
    int stat_result;
    striType owner;

  /* cmdGetOwner */
    logFunction(printf("cmdGetOwner(\"%s\")", striAsUnquotedCStri(filePath));
                fflush(stdout););
    os_path = cp_to_os_path(filePath, &path_info, &err_info);
    if (unlikely(os_path == NULL)) {
      logError(printf("cmdGetOwner: cp_to_os_path(\"%s\", *, *) failed:\n"
                      "path_info=%d, err_info=%d\n",
                      striAsUnquotedCStri(filePath), path_info, err_info););
      raise_error(err_info);
      owner = NULL;
    } else {
      stat_result = os_stat(os_path, &stat_buf);
      if (unlikely(stat_result != 0)) {
        logError(printf("cmdGetOwner: os_stat(\"" FMT_S_OS "\") failed:\n"
                        "errno=%d\nerror: %s\n",
                        os_path, errno, strerror(errno)););
        os_stri_free(os_path);
        raise_error(FILE_ERROR);
        owner = NULL;
      } else {
        os_stri_free(os_path);
        /* printf("cmdGetOwner: st_uid=" FMT_U32 "\n", stat_buf.st_uid); */
	owner = getUserFromUid(stat_buf.st_uid);
      } /* if */
    } /* if */
    logFunctionResult(printf("\"%s\"\n", striAsUnquotedCStri(owner)););
    return owner;
  } /* cmdGetOwner */



striType cmdUser (void)

  {
    striType user;

  /* cmdUser */
    user = getUserFromUid(getuid());
    logFunction(printf("cmdUser() --> \"%s\"", striAsUnquotedCStri(user)););
    return user;
  } /* cmdUser */
